/* functions for manipulating tree structures */

// global variables for tree plotting

var treeString="empty";
var treeRoot;
var spNames=[];
var maxNameLength=0;
var initX=40;
var initY=15;
var scaleFactor=50;
var spaceFactor=50
var heightFactor=100;
var heightToSpaceFactor=1;
var space=0;
var tallestTreeScale=false;
var useCladogram=false;
var treeVec = [];
var aString="";
var treePos="";
var maxHeight=0;
var value=0;

// var canvas; 
// var context;

    var canvas = document.getElementById('canvas');
    var context = canvas.getContext('2d');


/* code to generate random starting trees for control file  */

function minNode(name, left, right, father) {
   this.name = name;
   this.left = left;
   this.right = right;
   this.father = father;
}

// use array to represent tips of tree
function treeArray(tNode,tArray)
{
    if(tNode==null) return;
    var currNode=tNode;
    if(currNode.left != null)
        treeArray(currNode.left,tArray);
    if(currNode.right != null)
        treeArray(currNode.right,tArray);
    if((currNode.left == null)&&(currNode.right == null))
        tArray.push(currNode);
}

// generate a random tree for ntax = speciesNames.length
function randomTree(speciesNames)
{
var treeList = [];
tRoot = new minNode("root",null,null,null);
tRoot.left = new minNode("empty",null,null,tRoot);
tRoot.right = new minNode("empty",null,null,tRoot);
treeArray(tRoot,treeList);
while(treeList.length < speciesNames.length)
{
    x=Math.floor((Math.random() * treeList.length));
    treeList[x].left = new minNode("empty", null, null,treeList[x]);
    treeList[x].right = new minNode("empty", null, null,treeList[x]);
    treeList = [];
    treeArray(tRoot,treeList);
}
for(var i=0; i<treeList.length; i++)
    treeList[i].name=speciesNames[i];
return tRoot;
}

function recurseTree(tNode,newickSt)
{
    if(tNode==null) {
       return;
    }
    var currNode=tNode;
    if(currNode.left != null)
    {
        newickSt.newick += "(";
        recurseTree(currNode.left,newickSt);
     }
    if(currNode.right != null)
    {
        newickSt.newick += ", ";
        recurseTree(currNode.right,newickSt);
        newickSt.newick += ")";
    }
    if((currNode.left == null)&&(currNode.right == null))
        newickSt.newick += currNode.name;
}

function newickFromTree(tNode)
{
    treeSt={newick: ""};
    recurseTree(tNode,treeSt);
    return treeSt.newick;
}


// tree plotting functions

initialize();

function initialize() {
    // Register an event listener to                                                                                 
    // call the resizeCanvas() function each time                                                                    
    // the window is resized.                                                                                        
    window.addEventListener('resize', resizeCanvas, false);

    // Draw canvas border for the first time.                                                                        
    resizeCanvas();
}

function resizeCanvas() {
    context.font = "italic bold 24px serif";
    context.canvas.width  = window.innerWidth*0.80;
    context.canvas.height = window.innerHeight*0.7;
    drawOneTree(value,treeVec,useCladogram,context,tallestTreeScale,canvas.height*0.9-maxNameLength);
}

$( "#relscale" ).on("click",function() {
    if(tallestTreeScale)
    {
        tallestTreeScale=false;
                context.clearRect ( 0 , 0 , canvas.width, canvas.height );

        drawOneTree(value,treeVec,useCladogram,context,tallestTreeScale,canvas.height*0.9-maxNameLength);
    }
    else
    {
        tallestTreeScale=true;
                context.clearRect ( 0 , 0 , canvas.width, canvas.height );

        drawOneTree(value,treeVec,useCladogram,context,tallestTreeScale,canvas.height*0.9-maxNameLength);
    }
});

$( "#cladogram" ).on("click",function() {
    if(useCladogram)
    {
        useCladogram=false;
                context.clearRect ( 0 , 0 , canvas.width, canvas.height );

        drawOneTree(value,treeVec,useCladogram,context,tallestTreeScale,canvas.height*0.9-maxNameLength);
    }
    else
    {
        useCladogram=true;
                context.clearRect ( 0 , 0 , canvas.width, canvas.height );

        drawOneTree(value,treeVec,useCladogram,context,tallestTreeScale,canvas.height*0.9-maxNameLength);
    }
});

$( "#refreshtrees" ).on("click",function() {
    makeTrees();
});

function getMaxHeight(noTr,treeVec){
  maxHeight=0;
  for(var y=0; y<noTr; y++)
    {

      var h1=getTreeHeight(treeVec[y]);
      if(h1 > maxHeight)
        maxHeight = h1;
    }
}

function getTreeHeight(tree)
{
// strip all thetas out of tree
var Newtree = tree.replace(/(#\d.\d+)([eE](\+|-)?[0-9]+)?/g,"");
// count opening parenthese until first tip taxa is encountered
var p=0;
while(Newtree[p]=="(")
  p++;
// strip all leading "("s out of tree
Newtree = Newtree.replace(/\(+/,"");
// put elements of tree into a vector
var newick=Newtree.match(/((\+|-)?([0-9]+\.[0-9]*|\.[0-9]+)([eE](\+|-)?[0-9]+)?)|(\w+)|(\()|(\))|(\,)/g); // ([eE][-+]?[0-9]+)?) /g);
// get tree height
var blsum=0;
blsum += Number(newick[1]); j=0;
for(var k=2; k<newick.length; k++)
{
  if(newick[k]=="(")
    j++;
  else if ((newick[k]==")")&&(j==0)&&(p>1)) {
      blsum += Number(newick[k+1]); p--;
  }
  else if ((newick[k]==")")&&(j>0)) {
    j--;
  }
}
return blsum;
}


function getTreeFile(){
    context.textAlign = 'center';

    $('#treeInput').change(function () {
	context.clearRect ( 0 , 0 , canvas.width, canvas.height );
	treeInFile=$('#treeInput').val();
	$( "#refreshtrees" ).removeClass('not');
	makeTrees();
    });
}

function makeTrees(){
    var scaleBar=0.0;
    var treeCt=1;
    var aString = fs.readFileSync(treeInFile).toString();
    treeVec = aString.split("\n");
    if(treeVec[treeVec.length-2].match(";")==null)
          treeVec.pop();
    var noTr=treeVec.length-2;
    getMaxHeight(noTr,treeVec);
    $( "#slider" ).slider(
          {orientation: "horizontal",
         max: treeVec.length-2,
         min: 0,
         // value: 0,
         slide: function( event, ui ) {
            // While sliding, update the value in the #amount div element
            treePos = "Tree: " + (ui.value+1).toString() + "/" + (treeVec.length - 1).toString();
            $( "#amount" ).html( treePos );
             value = ui.value;
             context.clearRect ( 0 , 0 , canvas.width, canvas.height );
             drawOneTree(value,treeVec,useCladogram,context,tallestTreeScale,canvas.height*0.9-maxNameLength);}
          });    
    value = $( "#slider" ).slider( "value" );
    treePos = "Tree: " + (value+1).toString() + "/" + (treeVec.length - 1).toString();
    $( "#amount" ).html( treePos );
    context.clearRect ( 0 , 0 , canvas.width, canvas.height );
    treeFromNewick(treeVec[value],true);
    drawOneTree(value,treeVec,useCladogram,context,tallestTreeScale,canvas.height*0.9-maxNameLength); 
}

function drawOneTree(value,treeVec,useCladogram,context,tallestTreeScale,hF)
   {   
    if(value < treeVec.length)
       {
       if(!useCladogram)
           treeFromNewick(treeVec[value],true);
       else
           treeFromNewick(treeVec[value],false);
       spaceFactor = (canvas.width-initX)*0.9/spNames.length;
       space=0;
       if(!useCladogram)
       {
         if(tallestTreeScale)
         {
              heightFactor=hF/maxHeight;
         }
         else
        {
           heightFactor=hF/treeRoot.left.height;
        }
      }
      else {
        heightFactor=hF/treeRoot.height;
      }
       // draw scale bar at left
       // if(value==0)
       if(!useCladogram)
       {
         scaleBar=30.0/heightFactor;
         makeEdge(initX-40,initY+maxNameLength,initY+maxNameLength+scaleBar*heightFactor,context)
         context.fillText(scaleBar.toPrecision(1),initX-35,initY+maxNameLength+scaleBar*heightFactor);
       }
           context.font = "italic bold 16px serif";
	   printNames(spNames,context);
	   if(!useCladogram)
               postOrder(treeRoot,context,true);
           else
               postOrder(treeRoot,context,false);
       }
   }

getTreeFile();

function makeEdge(x,y,z,context)
    {
        context.lineWidth = 3;
        context.lineJoin = 'round';
        context.beginPath();
        context.moveTo(x,y);
        context.lineTo(x,z);
        context.stroke();
    }

function Node(data, left, right, father) {
   this.data = data;
   this.left = left;
   this.right = right;
   this.father = father;
   this.space = 0;
   this.height = 0;
   this.show = show;
}
function show() {
   return this.data;
}

function printNames(nameArray,context)
{
    var currX;
    var cY;
    currX=initX+spaceFactor;
    cY=initY+maxNameLength -12 ;
    for(var i=0; i<nameArray.length; i++)
    {
	context.textAlign='start';
	context.textBaseline='middle';
	context.save();
	context.translate(currX,cY);
	context.rotate(Math.PI*(3/2));	
        context.fillText(nameArray[i],0,0);
        // context.strokeText(nameArray[i],currX,cY);
	context.restore();
	currX+=spaceFactor;
    }
}

function postOrder(node,context,brLng)
{
    if(!brLng)  // is a cladogram with no branch lengths
    {
	if(node == null) return;
	if(node.left != null)
            postOrder(node.left,context,false);
	if(node.right != null)
            postOrder(node.right,context,false);
	if((node.left == null)&&(node.right == null))
	{
            space+=spaceFactor;
            node.space=space;
            makeEdge(node.space+initX,node.height*heightFactor+initY+maxNameLength,node.father.height*heightFactor+initY+maxNameLength,context);
	}
	else
	{
	    node.space = (node.left.space + node.right.space)/2;
	    context.lineWidth = 2;
	    context.lineJoin = 'round';
	    context.beginPath();
	    context.moveTo(node.left.space+initX,node.height*heightFactor+initY+maxNameLength);
	    context.lineTo(node.right.space+initX,node.height*heightFactor+initY+maxNameLength);
	    context.stroke();
	    if(node.father != null)
	    {
		var y1=node.right.height*heightFactor+initY+maxNameLength;
		var y2=node.height*heightFactor+initY+maxNameLength;
		makeEdge(node.space+initX,node.height*heightFactor+initY+maxNameLength,node.father.height*heightFactor+initY+maxNameLength,context);
            }
	}
    }
    else
	if(brLng)  // is an ultrametric tree with branch lengths
        {
	    if(node == null) return;
	    if(node.left != null)
		postOrder(node.left,context,true);
	    if(node.right != null)
		postOrder(node.right,context,true);
	    if((node.left == null)&&(node.right == null))
	    {
		space+=spaceFactor;
		node.space=space;
		var y2=node.height*heightFactor+initY+maxNameLength;
		makeEdge(node.space+initX,initY+maxNameLength,node.height*heightFactor+initY+maxNameLength,context);
	    }
	    else
	    {
		node.space = (node.left.space + node.right.space)/2;
		context.lineWidth = 2;
		context.lineJoin = 'round';
		context.beginPath();
		context.moveTo(node.left.space+initX,node.right.height*heightFactor+initY+maxNameLength);
		context.lineTo(node.right.space+initX,node.right.height*heightFactor+initY+maxNameLength);
		context.stroke();		
		if(node.father != null)
		{
		    var y1=node.right.height*heightFactor+initY+maxNameLength;
		    var y2=node.height*heightFactor+initY+maxNameLength;
		    makeEdge(node.space+initX,node.right.height*heightFactor+initY+maxNameLength,node.height*heightFactor+initY+maxNameLength,context);
		}
	    }
	}
}


function treeFromNewick(newickString,brLen)
{
    if(!brLen)
    {
	height = newickString.match(/(\,)/g).length;
	var newickString=newickString.replace(/(#\d+\.\d+)|(\d+\.\d+)/g,"").replace(/e-\d+/g,"").replace(/:/g,"");
	spNames = newickString.match(/(?=\D)(\w+)/g);
	getMaxLenSN(spNames);
	var newick=newickString.match(/(\w+)|(\()|(\))|(\,)/g);
	var n = new Node("root", null, null,null,null,height);
	treeRoot = n;
	current = treeRoot;
	current.height=0;
	for(var pos = 0; pos < newick.length; pos++)
        {
            if((newick[pos] == "(")||(newick[pos]==","))
                var n = new Node("empty", null, null,null,null,null);
            switch(newick[pos]) {
            case "(":
                // up left
                {
                    current.left = n;
                    n.father = current;
                    current = n;
                }
                break;
            case ",":
                // back then right
                {
                    current = current.father;
                    current.right=n;
                    n.father = current;
                    current = n;
                }
                break;
            case ")":
                // back
                {
                    current = current.father;
                    current.height=Math.max(current.right.height,current.left.height)+1;
                }
                break;
            default:
                {
                    current.data = newick[pos];
                    current.height = 0;
                }
		break;
            }
	}
	treeRoot.height=Math.max(current.right.height,current.left.height)+1;
    }
    else
	if(brLen)
        {
	    spNames = newickString.replace(/(#\d+\.\d+)|(\d+\.\d+)/g,"").replace(/e-\d+/g,"").replace(/:/g,"").match(/(?=\D)(\w+)/g);
	    getMaxLenSN(spNames);
	    var newickString=newickString.replace(/(#\d+\.\d+)([eE](\+|-)?[0-9]+)/g,"").replace(/:/g,"");
	    var newick=newickString.match(/((\+|-)?([0-9]+\.[0-9]*|\.[0-9]+)([eE](\+|-)?[0-9]+)?)|(\w+)|(\()|(\))|(\,)/g); 
	    var n = new Node("root", null, null,null,null,null);
	    treeRoot = n;
	    current = treeRoot;
	    cumY=0.0;
	    for(var pos = 0; pos < newick.length; pos++)
            {
		if(newick[pos] != ")")
                    var n = new Node("empty", null, null,null,null,null);
		switch(newick[pos]) {
		case "(":
                    // up left
                    {
			current.left = n;
			n.father = current;
			current = n;
                    }
                    break;
		case ",":
                    // back then right
                    {
			current = current.father;
			current.right=n;
			n.father = current;
			current = n;
                    }
                    break;
		case ")":
                    // back
                    {
			cumY = current.height;
			current = current.father;
                    }
                    break;
		case ";":
                    // at end
                    break;
		default:
                    {
			if(newick[pos].match(/(\+|-)?([0-9]+\.[0-9]*|\.[0-9]+)([eE](\+|-)?[0-9]+)?/) != null)
			{
                            current.data = newick[pos];
                            current.height = parseFloat(newick[pos])+cumY;
			}
			else
			{
			    cumY=0.0;
			}
                    }
		    break;
		}
            }
	}
}
    
function getMaxLenSN(sN)
{
    mLen=0;
    iD=0;
    for(var i=0; i<sN.length; i++)
	if(sN[i].length>mLen) { mLen=sN[i].length; iD=i; }
    context.font = "italic bold 16px serif";
    maxNameLength=context.measureText(sN[iD]).width;    
}



$( "#tabs" ).tabs();

$( "#dialog" ).dialog({
	autoOpen: false,
	width: 400,
	buttons: [
		{
			text: "Ok",
			click: function() {
				$( this ).dialog( "close" );
			}
		},
		{
			text: "Cancel",
			click: function() {
				$( this ).dialog( "close" );
			}
		}
	]
});
